/* Andrey N. Kuznetsov, Almaty, 2012 *
 * For educational purposes.         *
 * Under GPL 2.0                     */
//---------------------------------------------------------------------
#include "TCentralComponent.h"
//---------------------------------------------------------------------
#define tr(s) String::fromUTF8(s)
//---------------------------------------------------------------------
TCentralComponent::TCentralComponent() : Component("Central Component"),
                                         pNameLabel(0),
                                         pChooseButton(0),
                                         PlayThread("PlayThread")
{
    pNameLabel = new Label(L"Name Label", tr("Выберите файл..."));
    pNameLabel->setFont(Font(15.0000f, Font::bold));
    pNameLabel->setJustificationType(Justification::centred);
    pNameLabel->setEditable(false, false, false);
    pNameLabel->setColour(Label::textColourId, Colours::black);
    pNameLabel->setColour(Label::backgroundColourId, Colours::azure);
    pNameLabel->setColour(Label::outlineColourId, Colours::black);
    addAndMakeVisible(pNameLabel);

    pChooseButton = new TextButton(L"Choose Button");
    pChooseButton->setButtonText(tr("Выбрать..."));
    pChooseButton->addListener(this);
    addAndMakeVisible(pChooseButton);

    pPlayButton = new TextButton(L"Play Button");
    pPlayButton->setButtonText(tr("Играть"));
    pPlayButton->setColour(TextButton::buttonColourId, Colours::lightgreen);
    pPlayButton->addListener(this);
    addAndMakeVisible(pPlayButton);

    // Файл для воспроизведения ещё не открыт, аудио-поток пока отсутствует
    pCurrentAudioFileSource = 0;

    CurrentFile = File::getCurrentWorkingDirectory();

    // Создаём пустой менеджер аудио-устройств
    pAudioDeviceManager = new AudioDeviceManager();
    // Инициируем менеджер аудио-устройств
    String sError = pAudioDeviceManager->initialise(1, 2, 0, true);
    // Сохраняем текущее аудио-устройство в переменной
    pAudioDevice = pAudioDeviceManager->getCurrentAudioDevice();
    // Если устройство отсутствует, выводим сообщение об ошибке
    if(!sError.isEmpty()) 
    {
        AlertWindow::showMessageBox(
                                    AlertWindow::WarningIcon, 
                                    tr("Ошибка воспроизведения"), 
                                    (tr("Не удалось открыть аудио-устройство:\n\n") += sError), 
                                    tr("Принять"), 0
                                   );
    }
    // Устанавливаем связи между менеджером устройств, 
    // AudioSourcePlayer и AudioTransportSource
    pAudioDeviceManager->addAudioCallback(& AudioPlayer);
    TransportSource.addChangeListener(this);
    AudioPlayer.setSource(& TransportSource);

    // Если есть доступное аудио-устройство...
    if(pAudioDevice != 0) 
    {
        // ...выводим его название на ярлык программы
        String sName = pAudioDevice->getName();
        pNameLabel->setText(sName, false);
    }
    else 
    {
        pNameLabel->setText(tr("Нет устройства!"), false);
    }

    setSize (400, 150);

    // Создаём компонент для выбора файла
    pFileNameDialog = new FilenameComponent(
                                            String::empty,
                                            CurrentFile,
                                            true,
                                            false,
                                            false,
                                            "*.wav;*.aif;*.aiff;*.ogg;*.flac;*.mp3;*.wma",
                                            String::empty,
                                            tr("Выберите файл...")
                                           );
    File SearchDir = SearchDir.getSpecialLocation(File::userMusicDirectory);
    pFileNameDialog->setDefaultBrowseTarget(SearchDir);
    // Задаём его размер (иначе его не будет видно)
    pFileNameDialog->setSize(250, 25);

    PlayThread.startThread(3);
}
//---------------------------------------------------------------------
TCentralComponent::~TCentralComponent()
{
    PlayThread.signalThreadShouldExit();
    PlayThread.waitForThreadToExit(100);
    pAudioDeviceManager->closeAudioDevice();
    TransportSource.setSource(0);
    AudioPlayer.setSource(0);
    pAudioDeviceManager->removeAudioCallback(& AudioPlayer);
    if(pAudioDeviceManager) delete pAudioDeviceManager;
    if(pCurrentAudioFileSource) delete pCurrentAudioFileSource;
    if(pFileNameDialog) delete pFileNameDialog;
    deleteAllChildren();
}
//---------------------------------------------------------------------
void TCentralComponent::paint(Graphics& Canvas)
{
    Canvas.fillAll(Colours::azure);
}
//---------------------------------------------------------------------
void TCentralComponent::resized()
{
    pNameLabel->setBounds(10, 10, getWidth() - 20, 25);
    pChooseButton->setBounds(proportionOfWidth(0.5000f) - ((150) / 2), proportionOfHeight(0.5000f) - ((25) / 2), 150, 25);
    pPlayButton->setBounds(proportionOfWidth(0.5000f) - ((150) / 2), getHeight() - 35, 150, 25);
}
//---------------------------------------------------------------------
void TCentralComponent::buttonClicked(Button* pButton)
{
    if(pButton == pChooseButton)
    {
        // Создаём "обёртку" для показа диалога...
        DialogWindow::showModalDialog(tr("Выбор файла"),
                                      pFileNameDialog,
                                      NULL,
                                      Colours::azure,
                                      true,
                                      false,
                                      false);
 
        // Когда файл открыт...
        CurrentFile = pFileNameDialog->getCurrentFile();
        // проверяем, существует ли он
        if(!(CurrentFile == File::nonexistent)) 
        {
            // Выводим на ярлык название файла
            pNameLabel->setText(CurrentFile.getFileName(), false);

            // Если ранее осуществлялось воспроизведение - останавливаем его.
            TransportSource.stop();
            TransportSource.setSource(0);
            if(pCurrentAudioFileSource != 0) delete pCurrentAudioFileSource;
            AudioFormatManager FormatManager;
            // Загрузка файла в поток
            AudioFormatReader* pAudioFormatReader;
            // Регистрируем базовые форматы аудио
            FormatManager.registerBasicFormats();
            // Создаём чтеца для текущего аудио-файла
            pAudioFormatReader = FormatManager.createReaderFor(CurrentFile);

            if(pAudioFormatReader != 0)
            {
                // Если чтец существует, создаём с его помощью новый источник аудио-данных
                pCurrentAudioFileSource = new AudioFormatReaderSource(pAudioFormatReader, true);
                // и передаём их в AudioTransportSource
                TransportSource.setSource(pCurrentAudioFileSource, 32768, & PlayThread, pAudioFormatReader->sampleRate);
                if(pCurrentAudioFileSource == 0)
                {
                    AlertWindow::showMessageBox(AlertWindow::WarningIcon, tr("Ошибка воспроизведения"), tr("Отсутствует источник воспроизведения аудио-файла."), tr("Принять"), 0);
                    return;
                }
            }
            else AlertWindow::showMessageBox(
                                             AlertWindow::WarningIcon, 
                                             tr("Ошибка воспроизведения"), 
                                             tr("Не инициирован источник чтения аудио-файла.\nВозможно, файл повреждён."), 
                                             tr("Принять"), 0
                                            );
        }
        else
        {
            pNameLabel->setText(tr("Файл не выбран!"), false);
        }
    }
    else if(pButton == pPlayButton)
    {
        // Если идёт воспроизведение...
        if(TransportSource.isPlaying())
        {
            // останавливаем его
            TransportSource.stop();
        }
        else
        {
            // Иначе - устанавливаем позицию в начало потока воспроизведения...
            TransportSource.setPosition(0);
            // и начинаем проигрывание файла
            TransportSource.start();
        }
    }
}
//---------------------------------------------------------------------
void TCentralComponent::changeListenerCallback(ChangeBroadcaster*)
{
    // Если идёт воспроизведение...
    if(TransportSource.isPlaying())
    {
        pPlayButton->setButtonText(tr("Остановить"));
        // изменяем цвет кнопки на розовый
        pPlayButton->setColour(TextButton::buttonColourId, Colours::pink);
    }
    else
    {
        pPlayButton->setButtonText(tr("Играть"));
        // иначе - меняем на зелёный.
        pPlayButton->setColour(TextButton::buttonColourId, Colours::lightgreen);
    }
}
//---------------------------------------------------------------------

